# distutils: language = c
# cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True, infer_types=True

from libc.stdint cimport uint16_t, uint8_t, uint64_t
from libc.stddef cimport size_t
from libc.stdlib cimport malloc, free
from cpython.bytes cimport PyBytes_AsString, PyBytes_AsStringAndSize
from libc.stdlib cimport free
from cpython.unicode cimport PyUnicode_FromStringAndSize
from libc.string cimport strlen

cdef inline char* encode(str r):
    cdef:
        bytes b = r.encode('utf-8')
        char *data = PyBytes_AsString(b)
    return data

cpdef bint validate_rut(str input_rut, bint suspicious=False):
    """
    Valida un solo RUT (str) y devuelve True/False.
    """
    cdef char* data = encode(input_rut)
    cdef bint result = _validate_rut(data, suspicious)

    return result


cpdef validate_rut_string(str v, bint suspicious=False):
    cdef:
        char* data = encode(v)
        bint result = _validate_rut(data, suspicious)
    if result:
        return v
    else:
        raise ValueError(f"Invalid RUT: {v}")


cpdef str format_rut(str rut, bint dots=True, bint uppercase=True, bint ignore_invalid=False):
    """
    Formatea un RUT (str) con puntos y guión.
    """
    cdef bytes b = rut.encode('utf-8')
    cdef char *data = PyBytes_AsString(b)
    cdef char *formatted_rut = format_rut_c(data, dots, uppercase, ignore_invalid)
    cdef str result = formatted_rut.decode('utf-8')
    free(formatted_rut)
    return result


cdef char* format_rut_c(const char* rut,
                        bint dots,
                        bint uppercase,
                        bint ignore_invalid):
    cdef const char* p = rut
    cdef char c, dv_char = 0
    cdef Py_ssize_t total_valid = 0
    cdef size_t i, body_len, out_len
    cdef char* out
    cdef char* dst
    cdef Py_ssize_t ndots = 0

    # ——————————————
    # 1) Primer pase: contar dígitos + capturar DV
    # ——————————————
    for i in range(strlen(rut)):
        c = p[i]
        if b'0' <= c <= b'9' or c == b'K' or c == b'k':
            dv_char = c
            total_valid += 1
        elif c == b'.' or c == b'-' or c == b' ':
            continue
        else:
            if not ignore_invalid:
                return NULL

    if total_valid < 2:
        return NULL

    body_len = total_valid - 1

    if uppercase:
        dv_char &= 0xDF  # convierte 'k' a 'K' si procede

    if dots:
        ndots = (body_len - 1) // 3

    # body_len dígitos + ndots puntos + '-' + DV + '\0'
    out_len = body_len + ndots + 3

    out = <char*> malloc(out_len)
    if not out:
        return NULL

    # ——————————————
    # 3) Segundo pase: construir la cadena formateada
    # ——————————————
    dst = out
    total_valid = 0  # reusar como contador de dígitos emitidos

    for i in range(strlen(rut)):
        c = p[i]
        if b'0' <= c <= b'9' or c == b'K' or c == b'k':
            if total_valid < body_len:
                # insertar punto si toca
                if dots and total_valid > 0 and ((body_len - total_valid) % 3) == 0:
                    dst[0] = b'.'; dst += 1

                # emitir dígito o 'K'/'k'
                if (c == b'K' or c == b'k'):
                    if uppercase:
                        dst[0] = b'K'
                    else:
                        dst[0] = b'k'
                else:
                    dst[0] = c
                dst += 1

            total_valid += 1

    # añadir guión, dígito verificador y terminador
    dst[0] = b'-'; dst += 1
    dst[0] = dv_char; dst += 1
    dst[0] = b'\0'

    return out


cdef inline uint8_t compute_dv_from_int(uint64_t body) nogil:
    """
    Calcula el dígito verificador de un RUT dado como entero (sin DV).
    Devuelve:
        - 0–9 para DV '0'–'9'
        - 10 para DV 'K'
    """
    cdef uint64_t n = body
    cdef int total = 0
    cdef int mul = 2
    cdef int digit

    # Recorrer dígitos de derecha a izquierda
    while n > 0:
        digit = <int>(n % 10)
        total += digit * mul
        mul = mul + 1 if mul < 7 else 2
        n //= 10

    cdef int m = 11 - (total % 11)
    if m == 11:
        return 0    # equivale a '0'
    elif m == 10:
        return 10   # equivale a 'K'
    else:
        return m

cdef inline str _verification_digit_from_int(int dv):
    cdef char dv_char
    if dv == 11:
        dv_char = b'0'
    elif dv == 10:
        dv_char = b'K'
    else:
        dv_char = <char>(dv + 0)  # 48 == ord('0')


    return PyUnicode_FromStringAndSize(&dv_char, 1)


cpdef str verification_digit(rut):
    """
    Devuelve el dígito verificador:
        - 0..9 para '0'..'9'
        - 10 para 'K'
    Acepta `rut` como str (sin DV) o como int (sin DV).
    """
    cdef char *data
    cdef Py_ssize_t length
    cdef int dv
    cdef bytes b

    if isinstance(rut, str):
        # 1) codifica a bytes y extrae buffer + longitud
        b = rut.encode('utf-8')
        if PyBytes_AsStringAndSize(b, &data, &length) != 0:
            raise ValueError("Error al codificar RUT a bytes")
        # 2) calcula DV sobre la cadena
        dv = compute_dv(data, <size_t>length)

        return _verification_digit_from_int(dv)

    elif isinstance(rut, int):
        # 3) convierte el entero a uint64_t y calcula DV
        dv = compute_dv_from_int(<uint64_t>rut)

        return _verification_digit_from_int(dv)

    else:
        raise TypeError(f"Tipo no válido: {type(rut).__name__}")

cdef inline char compute_dv(char *s, size_t body_len) nogil:
    cdef uint16_t total = 0
    cdef uint16_t m
    cdef int mul = 2
    cdef size_t i = body_len
    # recorrer de derecha a izquierda sin usar range
    while i > 0:
        i -= 1
        total += (s[i] - 48) * mul  # '0' -> 48
        if mul < 7:
            mul += 1
        else:
            mul = 2
    m = 11 - (total % 11)
    if m == 11:
        return <char>48  # '0'
    elif m == 10:
        return <char>75  # 'K'
    else:
        return <char>(48 + m)

cdef inline void clean_rut(const char* src, char* dst) nogil:
    cdef char c
    while True:
        c = src[0]
        if c == b'\0':
            break
        src += 1
        if c >= b'0' and c <= b'9':
            dst[0] = c
            dst += 1
        elif c == b'K' or c == b'k':
            dst[0] = b'K'
            dst += 1
        elif c == b'.' or c == b'-' or c == b' ':
            continue
        else:
            # carácter inválido, limpiar salida
            dst[0] = b'\0'
            return
    dst[0] = b'\0'


cdef inline bint is_suspicious(char* s) nogil:
    """
    Devuelve True si todos los caracteres en s (hasta '\0')
    son iguales entre sí. Si la cadena está vacía (s[0] == '\0'),
    se considera que cumple (devuelve True).
    """
    cdef char first = s[0]
    cdef char c

    # Si la cadena está vacía, asumimos que "todos los caracteres son iguales"
    if first == b'\0':
        return True

    # Avanzamos un puntero para comparar el resto
    s += 1
    while True:
        c = s[0]
        if c == b'\0':
            break
        if c != first:
            return False
        s += 1

    return True



cdef bint _validate_rut(char *s, bint suspicious):
    cdef const char *src = s
    cdef char cleaned[99]  # Buffer de salida, tamaño suficiente para un RUT
    clean_rut(src, cleaned)

    # Verifica si el resultado quedó vacío (fallo por carácter inválido)
    if cleaned[0] == b'\0':
        return False

    # Longitud del RUT limpio
    cdef size_t length = strlen(cleaned)
    if length < 2:
        return False

    if suspicious:
        if is_suspicious(cleaned):
            return False

    cdef char expected = compute_dv(cleaned, length - 1)

    return cleaned[length - 1] == expected



import pytest
import pyrut
from ..pyrut.types import Rut, RutNotSuspicious
from pydantic import ValidationError


class TestValidateRut:
    """Tests for validate_rut function"""

    def test_valid_ruts(self):
        """Test with valid RUTs in different formats"""
        valid_ruts = [
            "11.111.111-1",
            "11111111-1",
            "111111111",
            "12.345.678-5",
            "12345678-5",
            "123456785",
            "7.654.321-K",
            "7654321-K",
            "7654321K",
            "7654321k",
            "1.234.567-0",
            "12345670",
            "9.999.999-9",
            "99999999"
        ]

        for rut in valid_ruts:
            assert pyrut.validate_rut(rut), f"RUT {rut} should be valid"

    def test_invalid_ruts(self):
        """Test with invalid RUTs"""
        invalid_ruts = [
            "11.111.111-2",  # Wrong check digit
            "12.345.678-6",  # Wrong check digit
            "7.654.321-J",   # Invalid check digit
            "123456789",     # Wrong length
            "1234567",       # Too short
            "12345678901",   # Too long
            "",              # Empty string
            "abcdefgh-i",    # Non-numeric body
            "12.34x.678-5",  # Invalid character in body
            "12345678-Z",    # Invalid check digit
            "12345678--5",   # Double dash
            "12..345.678-5", # Double dot
        ]

        for rut in invalid_ruts:
            assert not pyrut.validate_rut(rut), f"RUT {rut} should be invalid"

    def test_suspicious_ruts(self):
        """Test suspicious RUTs (all same digits)"""
        suspicious_ruts = [
            "11111111-1",
            "22222222-2",
            "33333333-3",
            "44444444-4",
            "55555555-5",
            "66666666-6",
            "77777777-7",
            "88888888-8",
            "99999999-9"
        ]

        # Without suspicious flag, they should be valid
        for rut in suspicious_ruts:
            assert pyrut.validate_rut(rut, suspicious=False)

        # With suspicious flag, they should be invalid
        for rut in suspicious_ruts:
            assert not pyrut.validate_rut(rut, suspicious=True)

    def test_edge_cases(self):
        """Test edge cases"""
        # Minimum valid RUT
        assert pyrut.validate_rut("1-9")

        # RUT with spaces
        assert pyrut.validate_rut(" 12.345.678-5 ")

        # Mixed case K
        assert pyrut.validate_rut("7654321k") == pyrut.validate_rut("7654321K")


class TestFormatRut:
    """Tests for format_rut function"""

    def test_format_with_dots(self):
        """Test formatting with dots enabled"""
        test_cases = [
            ("12345678-5", "12.345.678-5"),
            ("123456785", "12.345.678-5"),
            ("7654321K", "7.654.321-K"),
            ("7654321k", "7.654.321-K"),
            ("1234567-0", "1.234.567-0"),
            ("12345-6", "12.345-6"),
            ("123-4", "123-4"),
            ("12-3", "12-3"),
            ("1-9", "1-9")
        ]

        for input_rut, expected in test_cases:
            result = pyrut.format_rut(input_rut, dots=True, uppercase=True)
            assert result == expected, f"Format {input_rut} -> expected {expected}, got {result}"

    def test_format_without_dots(self):
        """Test formatting without dots"""
        test_cases = [
            ("12.345.678-5", "12345678-5"),
            ("123456785", "12345678-5"),
            ("7.654.321-K", "7654321-K"),
            ("7654321k", "7654321-K")
        ]

        for input_rut, expected in test_cases:
            result = pyrut.format_rut(input_rut, dots=False, uppercase=True)
            assert result == expected, f"Format {input_rut} -> expected {expected}, got {result}"

    def test_format_lowercase(self):
        """Test formatting with lowercase K"""
        test_cases = [
            ("7654321K", "7.654.321-k"),
            ("7654321k", "7.654.321-k"),
            ("7.654.321-K", "7.654.321-k")
        ]

        for input_rut, expected in test_cases:
            result = pyrut.format_rut(input_rut, dots=True, uppercase=False)
            assert result == expected, f"Format {input_rut} -> expected {expected}, got {result}"

    def test_format_combinations(self):
        """Test all combinations of formatting options"""
        input_rut = "7654321k"

        # dots=True, uppercase=True
        assert pyrut.format_rut(input_rut, dots=True, uppercase=True) == "7.654.321-K"

        # dots=True, uppercase=False
        assert pyrut.format_rut(input_rut, dots=True, uppercase=False) == "7.654.321-k"

        # dots=False, uppercase=True
        assert pyrut.format_rut(input_rut, dots=False, uppercase=True) == "7654321-K"

        # dots=False, uppercase=False
        assert pyrut.format_rut(input_rut, dots=False, uppercase=False) == "7654321-k"

    def test_format_ignore_invalid(self):
        """Test format with ignore_invalid flag"""
        # Invalid RUT should raise error by default
        with pytest.raises(Exception):
            pyrut.format_rut("invalid-rut", ignore_invalid=False)

        # With ignore_invalid=True, it might handle differently
        # This depends on the implementation details


class TestVerificationDigit:
    """Tests for verification_digit function"""

    def test_verification_digit_string(self):
        """Test verification digit calculation with string input"""
        test_cases = [
            ("12345678", "5"),
            ("11111111", "1"),
            ("7654321", "K"),
            ("1234567", "0"),
            ("9999999", "9"),
            ("1", "9"),
            ("12", "3"),
            ("123", "4")
        ]

        for rut_body, expected_dv in test_cases:
            result = pyrut.verification_digit(rut_body)
            assert result == expected_dv, f"VD for {rut_body} -> expected {expected_dv}, got {result}"

    def test_verification_digit_int(self):
        """Test verification digit calculation with integer input"""
        test_cases = [
            (12345678, "5"),
            (11111111, "1"),
            (7654321, "K"),
            (1234567, "0"),
            (9999999, "9"),
            (1, "9"),
            (12, "3"),
            (123, "4")
        ]

        for rut_body, expected_dv in test_cases:
            result = pyrut.verification_digit(rut_body)
            assert result == expected_dv, f"VD for {rut_body} -> expected {expected_dv}, got {result}"

    def test_verification_digit_edge_cases(self):
        """Test edge cases for verification digit"""
        # Test with very large numbers
        large_num = 999999999
        dv = pyrut.verification_digit(large_num)
        assert isinstance(dv, str)
        assert dv in "0123456789K"

        # Test with zero
        assert pyrut.verification_digit(0) in "0123456789K"

    def test_verification_digit_invalid_input(self):
        """Test verification digit with invalid input types"""
        invalid_inputs = [
            [],
            {},
            None,
            3.14,
            "abc"
        ]

        for invalid_input in invalid_inputs:
            with pytest.raises((TypeError, ValueError)):
                pyrut.verification_digit(invalid_input)


class TestPydanticTypes:
    """Tests for Pydantic type annotations"""

    def test_rut_type_valid(self):
        """Test Rut type with valid RUTs"""
        from pydantic import BaseModel

        class TestModel(BaseModel):
            rut: Rut

        valid_ruts = [
            "11.111.111-1",
            "12345678-5",
            "7654321K"
        ]

        for rut in valid_ruts:
            model = TestModel(rut=rut)
            assert model.rut == rut

    def test_rut_type_invalid(self):
        """Test Rut type with invalid RUTs"""
        from pydantic import BaseModel

        class TestModel(BaseModel):
            rut: Rut

        invalid_ruts = [
            "11.111.111-2",
            "invalid-rut",
            "12345678-Z"
        ]

        for rut in invalid_ruts:
            with pytest.raises(ValidationError):
                TestModel(rut=rut)

    def test_rut_not_suspicious_type(self):
        """Test RutNotSuspicious type"""
        from pydantic import BaseModel

        class TestModel(BaseModel):
            rut: RutNotSuspicious

        # Valid non-suspicious RUT should work
        model = TestModel(rut="12.345.678-5")
        assert model.rut == "12.345.678-5"

        # Suspicious RUT should fail
        with pytest.raises(ValidationError):
            TestModel(rut="11.111.111-1")


class TestIntegration:
    """Integration tests combining multiple functions"""

    def test_format_and_validate_cycle(self):
        """Test that formatted RUTs are still valid"""
        test_ruts = [
            "12345678-5",
            "7654321K",
            "1234567-0"
        ]

        for rut in test_ruts:
            # Format the RUT
            formatted = pyrut.format_rut(rut, dots=True, uppercase=True)
            # Validate the formatted RUT
            assert pyrut.validate_rut(formatted), f"Formatted RUT {formatted} should be valid"

    def test_verification_digit_integration(self):
        """Test that calculated verification digits make valid RUTs"""
        test_bodies = ["12345678", "7654321", "1234567"]

        for body in test_bodies:
            dv = pyrut.verification_digit(body)
            full_rut = f"{body}-{dv}"
            assert pyrut.validate_rut(full_rut), f"RUT {full_rut} with calculated DV should be valid"

    def test_clean_and_validate(self):
        """Test validation with various dirty inputs"""
        dirty_ruts = [
            "  12.345.678-5  ",
            "12 345 678-5",
            "12.345.678 - 5",
            "12-345-678-5"
        ]

        for dirty_rut in dirty_ruts:
            # Should either validate or fail gracefully
            result = pyrut.validate_rut(dirty_rut)
            assert isinstance(result, bool)


class TestPerformance:
    """Performance and stress tests"""

    def test_validate_many_ruts(self):
        """Test validation of many RUTs"""
        # Generate test RUTs
        test_ruts = []
        for i in range(1000000, 1001000):  # 1000 RUTs
            body = str(i)
            dv = pyrut.verification_digit(body)
            test_ruts.append(f"{body}-{dv}")

        # Validate all RUTs
        for rut in test_ruts:
            assert pyrut.validate_rut(rut), f"Generated RUT {rut} should be valid"

    def test_format_many_ruts(self):
        """Test formatting of many RUTs"""
        test_ruts = [f"1234567{i}-{pyrut.verification_digit(f'1234567{i}')}"
                    for i in range(10)]

        for rut in test_ruts:
            formatted = pyrut.format_rut(rut, dots=True, uppercase=True)
            assert isinstance(formatted, str)
            assert len(formatted) > len(rut.replace(".", ""))  # Should be longer due to dots



from setuptools import setup, Extension, find_packages
from Cython.Build import cythonize
import pathlib

here = pathlib.Path(__file__).parent

extensions = [
    Extension(
        "pyrut",
        sources=["pyrut/rut.pyx"],
        extra_compile_args=["-O3", "-march=native"],
        language="c"
    ),
]

setup(
    name="PyRut",
    version="1.0.7",
    description="High-performance Chilean RUT validation & formatting (Cython)",
    long_description=here.joinpath("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    author="Bastian Garcia",
    author_email="bastiang@uc.cl",
    url="https://github.com/cve-zh00/PyRut",
    packages=find_packages(),
    ext_modules=cythonize(extensions),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Cython",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    zip_safe=False,
)
